#include <tcl-gobject.h>
#include <string.h>

/* A handle object represents a pointer to a GType.
 * It stores both the pointer and its GType and does type
 * checking when retrieving the pointer from a handle.
 * The string rep looks like "GtkButton 0x1234567".
 * The special string rep "null" represents the NULL pointer
 * and bypasses type checking.
 */

static void updateHandleString (Tcl_Obj *);
static int setHandleFromAny (Tcl_Interp *, Tcl_Obj *);

static Tcl_ObjType handleType = {
  "handle",
  0,
  0,
  updateHandleString,
  setHandleFromAny
};

void initHandleType ()
{
  Tcl_RegisterObjType (&handleType);
}

static int setHandleFromAny (Tcl_Interp *interp, Tcl_Obj *obj)
{
  void *ptr;
  char *s, *s2;
  GType type;

  s = Tcl_GetString (obj);

  if (strcmp (s, "null") == 0) {
      ptr = 0;
      type = G_TYPE_NONE;
  } else {
    s2 = g_new (char, strlen (s));

    if (sscanf(s, "%s %p", s2, &ptr) != 2) {
      goto error;
    }

    type = g_type_from_name (s2);

    if (type == G_TYPE_INVALID) {
      goto error;
    }

    g_free (s2);
  }

  if (obj->typePtr && obj->typePtr->freeIntRepProc) {
    obj->typePtr->freeIntRepProc (obj);
  }

  obj->typePtr = &handleType;
  obj->internalRep.twoPtrValue.ptr1 = ptr;
  obj->internalRep.twoPtrValue.ptr2 = GUINT_TO_POINTER (type);

  return TCL_OK;

 error:

  if (interp) {
    appendResult (interp, "bad handle: ", s, (char *) 0);
  }

  g_free (s2);
  
  return TCL_ERROR;
}

static void updateHandleString (Tcl_Obj *obj)
{
  const char *s;

  if (HANDLE_PTR (obj)) {
    s = g_type_name (HANDLE_GTYPE (obj));
    obj->length = strlen (s) + (sizeof(void *) * 3) + 1;
    obj->bytes = Tcl_Alloc (obj->length);
    obj->length = snprintf (obj->bytes, obj->length, "%s %p", s,
			    HANDLE_PTR (obj));
  } else {
    s = "null";
    obj->length = strlen (s);
    obj->bytes = Tcl_Alloc (obj->length);
    strcpy (obj->bytes, s);
  }
}

Tcl_Obj *newHandle (void *ptr, GType type)
{
  Tcl_Obj *obj;

  obj = Tcl_NewObj ();
  obj->bytes = 0;
  obj->typePtr = &handleType;
  obj->internalRep.twoPtrValue.ptr1 = ptr;

  if (ptr) {
    if (G_TYPE_FUNDAMENTAL (type) == G_TYPE_OBJECT) {      
      type = G_OBJECT_TYPE (ptr);
    }
  } else {
    type = G_TYPE_NONE;
  }

  obj->internalRep.twoPtrValue.ptr2 = GUINT_TO_POINTER (type);

  return obj;
}

int getHandle (Tcl_Interp *interp, Tcl_Obj *obj, GType type, void **ptr)
{
  if (obj->typePtr != &handleType && setHandleFromAny (interp, obj) !=TCL_OK) {
    return TCL_ERROR;
  }

  *ptr = HANDLE_PTR (obj);

  if (*ptr) {
    if (!g_type_is_a (HANDLE_GTYPE (obj), type)) {
      appendResult (interp, "wrong type: ", g_type_name (HANDLE_GTYPE (obj)),
		   ", must be a: ", g_type_name (type), (char *) 0);
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}

int cmd_handle (ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *CONST objv[])
{
  int i;
  Tcl_Obj *obj = objv[2];
  static const char *cmd[] = {"is_null", "get_type", (char *) 0};

  enum {
    CMD_IS_NULL, CMD_GET_TYPE
  };

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command option arg...");
    return TCL_ERROR;
  }

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "handle");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj (interp, objv[1], cmd, "cmd", TCL_EXACT, &i)
      != TCL_OK) {
    return TCL_ERROR;
  }

  if (obj->typePtr != &handleType &&
      setHandleFromAny (interp, obj) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (i) {
  case CMD_IS_NULL:
    Tcl_SetObjResult (interp, Tcl_NewBooleanObj (HANDLE_PTR (obj) == 0));
    break;
    
  case CMD_GET_TYPE:
    Tcl_SetObjResult (interp, Tcl_NewIntObj (HANDLE_GTYPE (obj)));
    break;

  default:
    g_assert_not_reached ();
  }
  
  return TCL_OK;
}
