/**!
 * gobject.notes:
 *
 * Only the useful parts of the GObject library are wrapped.
 * At some point it would be nice to add the ability to define new
 * class types in Tcl.
 * 
 * Functions that use the GError** will instead take the name of a
 * variable. If there's an error that variable will be set to a list
 * of the form '{domain_id error_id msg}'.
 *
 * See &{g::signal connect} for information on signal handling.
 **/

static int getPropertyGValue (Tcl_Interp *interp, GType type,
				const char *name, GValue *val)
{
  GParamSpec *info;

  info = g_object_class_find_property (type_class (type), name);
  
  if (!info) {
    appendResult (interp, "bad property name ", name, (char *) 0);
    return TCL_ERROR;
  }

  g_value_init (val, info->value_type);

  return TCL_OK;
}

static int getGSignalID (Tcl_Interp *interp, GType type, Tcl_Obj *obj,
			 guint *id)
{
  const char *name = Tcl_GetString (obj);

  *id = g_signal_lookup (name, type);

  if (!*id) {
    appendResult (interp, "bad signal name ", name, (char *) 0);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/**!
 * gobject.func.g::object new:
 * g::object new
 * g_object_newv
 *
 *> Usage
 *   @type         a %GType
 *   @arglist      a list of the form {name value ...}
 *   Return value  The new %GObject.
 *
 * Creates a %GObject of type @type with property values given
 * by @arglist. Wraps %g_object_newv.
 **/

static  int cmd_g_object_new (Tcl_Interp *interp, int objc,
			      Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GParameter *param;
  int i, j, len;
  Tcl_Obj **obj;

  /* G_TYPE_UINT */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "{name value ...}");

  if (!arg) {
    return TCL_ERROR;
  }

  if (Tcl_ListObjGetElements (interp, objv[3], &len, &obj) != TCL_OK) {
    return TCL_ERROR;
  }

  if (len & 1) {
    appendResult(interp, "must be {name value ...}", (char *) 0);
    return TCL_ERROR;
  }

  len /= 2;
  param = g_new0 (GParameter, len);
 
  for (i = 0, j = 0; j < len; j++) {
    param[j].name = Tcl_GetString (obj[i++]);
    
    if (getPropertyGValue (interp, arg[1].int_val, param[j].name,
			   &param[j].value) != TCL_OK) {
      g_free (param);
      return TCL_ERROR;
    }

    if (getGValue (interp, obj[i++], &param[j].value) != TCL_OK) {
      g_free (param);
      return TCL_ERROR;
    }
  }

  /* GObject* GType guint GParameter* */
  arg[0].ptr_val = g_object_newv (arg[1].int_val, len, param);

  /* G_TYPE_OBJECT */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  g_free (param);
  return TCL_OK;
}

/**!
 * gobject.func.g::object set:
 * g::object set
 * g_object_set_property
 * 
 *> Usage
 *   @obj     GObject*
 *   @name    const gchar*
 *   @value   any type
 *
 * Wraps %g_object_set_property.
 **/

static int cmd_g_object_set (Tcl_Interp *interp, int objc,
			     Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GValue val = {0};

  /* G_TYPE_OBJECT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 5, sig, 0X4, "val");

  if (!arg) {
    return TCL_ERROR;
  }

  if (getPropertyGValue (interp, G_OBJECT_TYPE (arg[1].ptr_val),
			 arg[2].ptr_val, &val) != TCL_OK) {
    return TCL_ERROR;
  }

  if (getGValue (interp, objv[4], &val) != TCL_OK) {
    return TCL_ERROR;
  }

  /* void GObject* {const gchar*} {const GValue*} */
  g_object_set_property (arg[1].ptr_val, arg[2].ptr_val, &val);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::object get:
 * g::object get
 * g_object_get_property
 * 
 *> Usage
 *   @obj           %GObject
 *   @name          string
 *   Return value   value of @name.
 *
 * Wraps %g_object_get_property.
 **/

static  int cmd_g_object_get (Tcl_Interp *interp, int objc,
			      Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GValue val = {0};

  /* G_TYPE_OBJECT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  if (getPropertyGValue (interp, G_OBJECT_TYPE (arg[1].ptr_val),
			 arg[2].ptr_val, &val) != TCL_OK) {
    return TCL_ERROR;
  }

  /* void GObject* {const gchar*} GValue* */
  g_object_get_property (arg[1].ptr_val, arg[2].ptr_val, &val);

  Tcl_SetObjResult (interp, newGValue (&val));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::object get_data:
 * g::object get_data
 * g_object_get_data
 * 
 *> Usage
 *   @obj          %GObject
 *   @name         string
 *   Return value  the data
 *
 * Wraps %g_object_get_data.
 **/


static int cmd_g_object_get_data (Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;

  /* G_TYPE_OBJECT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* gpointer GObject* {const gchar*} */
  arg[0].ptr_val = g_object_get_data (arg[1].ptr_val, arg[2].ptr_val);

  if (!arg[0].ptr_val) {
    appendResult(interp, "bad key", (char *) 0);
    return TCL_ERROR;
  }

  Tcl_SetObjResult (interp, ((cb_t *) arg[0].ptr_val)->data);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::object set_data:
 * g::object set_data
 * g_object_set_data
 * 
 *> Usage
 *   @obj    %GObject
 *   @name   string
 *   @data   any type
 *
 * Wraps %g_object_set_data.
 **/

static int cmd_g_object_set_data (Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;

  /* G_TYPE_OBJECT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 5, sig, 0X4, "data");

  if (!arg) {
    return TCL_ERROR;
  }

  /* void GObject* {const gchar*} gpointer */
  g_object_set_data_full (arg[1].ptr_val, arg[2].ptr_val,
			  cb_new (0, 0, objv[4]),
			  (GDestroyNotify) cb_free);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::object set_data_full:
 * g::object set_data_full
 * g_object_set_data_full
 * 
 *> Usage
 *   @obj     %GObject
 *   @name    string
 *   @value   any type
 *   @script  script
 *
 * Wraps %g_object_set_data_full. Set @name to @value in @obj and eval
 * @script when @name is removed.
 **/

static int cmd_g_object_set_data_full (Tcl_Interp *interp, int objc,
				Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;

  /* G_TYPE_OBJECT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 6, sig, 0XC, "data script");

  if (!arg) {
    return TCL_ERROR;
  }

  /* void GObject* {const gchar*} gpointer */
  g_object_set_data_full (arg[1].ptr_val, arg[2].ptr_val,
			  cb_new (interp, objv[5], objv[4]),
			  (GDestroyNotify) cb_exec_once);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::object steal_data:
 * g::object steal_data
 * g_object_steal_data
 *
 *> Usage
 *   @name          %GObject
 *   @name          string
 *   Return value   the value of @name
 *
 * Wraps %g_object_steal_data.
 **/

static int cmd_g_object_steal_data (Tcl_Interp *interp, int objc,
			     Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  
  /* G_TYPE_OBJECT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* gpointer GObject* {const gchar*} */
  arg[0].ptr_val = g_object_steal_data (arg[1].ptr_val, arg[2].ptr_val);

  if (!arg[0].ptr_val) {
    appendResult(interp, "bad key", (char *) 0);
    return TCL_ERROR;
  }

  Tcl_SetObjResult (interp, ((cb_t *) arg[0].ptr_val)->data);

  cb_free (arg[0].ptr_val);
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}


/**!
 * gobject.func.g::signal list:
 * g::signal list
 * g_signal_list_ids
 * 
 *> Usage
 *   @type          %GType
 *   Return value   List of signals names.
 *
 * Wraps %g_signal_list_ids.
 **/

static int cmd_g_signal_list (Tcl_Interp *interp, int objc,
			      Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  guint len;
  guint *id;
  Tcl_Obj *obj;
  cval_t *arg;

  /* G_TYPE_UINT */
  arg = cval_gather (interp, objc, objv, 3, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* guint* GType guint* */
  id = g_signal_list_ids (arg[1].int_val, &len);

  obj = Tcl_NewListObj (0, 0);
  
  while (len-- > 0) {
    Tcl_ListObjAppendElement (interp, obj,
			      Tcl_NewStringObj (g_signal_name (*id++), -1));
  }
  
  /* FIXME: free id? */

  Tcl_SetObjResult (interp, obj);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

static gboolean emission_hook (GSignalInvocationHint *ihint,
				guint n_param,
				const GValue *param, gpointer data)
{
  Tcl_Obj *arg, *ret;
  cb_t *cb = data;
  int i;

  arg = Tcl_NewListObj (0, 0);

  for (i = 0; i < n_param; i++) {
    Tcl_ListObjAppendElement (cb->interp, arg, newGValue (&param[i]));
  }

  Tcl_SetVar2Ex (cb->interp, "_", 0, arg, TCL_GLOBAL_ONLY);

  ret = cb_exec (cb);

  if (ret) {
    if (Tcl_GetBooleanFromObj (cb->interp, ret, &i) != TCL_OK) {
      Tcl_BackgroundError(cb->interp);
    }

    return i;
  }

  return TRUE;
}

/**!
 * gobject.func.g::signal add_emission_hook:
 * g::signal add_emission_hook
 * g_signal_add_emission_hook
 *
 *> Usage
 *   @type:         %GType
 *   @signal:       string
 *   @script:       script
 *   Return value   %gulong id
 *
 * Wraps %g_signal_add_emission_hook. The script is evaluated as
 * in &{g::signal connect}.
 **/

static int cmd_g_signal_add_emission_hook (Tcl_Interp *interp, int objc,
					   Tcl_Obj *CONST objv[],
					   const cmd_sig_t *sig)
{
  guint id, type;
  long hook_id;

  if (objc != 5) {
    Tcl_WrongNumArgs(interp, 2, objv, "type signal script");
    return TCL_ERROR; 
  }

  if (Tcl_GetIntFromObj (interp, objv[2], &type) != TCL_OK) {
    return TCL_ERROR;
  }

  if (getGSignalID (interp, type, objv[3], &id) != TCL_OK) {
    return TCL_ERROR;
  }

  /* gulong guint GQuark GSignalEmissionHook gpointer GDestroyNotify */
  hook_id = g_signal_add_emission_hook (id, 0, emission_hook,
					cb_new (0, objv[3], 0),
					(GDestroyNotify) cb_free);
  
  /* G_TYPE_ULONG */
  Tcl_SetObjResult (interp, Tcl_NewLongObj (hook_id));

  return TCL_OK;
}

/**!
 * gobject.func.g::signal remove_emission_hook:
 * g::signal remove_emission_hook
 * g_signal_remove_emission_hook
 * 
 *> Usage
 *   @signal   string
 *   @id       %gulong
 *
 * Wraps %g_signal_remove_emission_hook.
 **/

static int cmd_g_signal_remove_emission_hook (Tcl_Interp *interp, int objc,
					      Tcl_Obj *CONST objv[],
					      const cmd_sig_t *sig)
{
  cval_t *arg;
  guint id;
  GType type;

  /* G_TYPE_ULONG */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X3, "type signal");

  if (Tcl_GetLongFromObj (interp, objv[2], &type) != TCL_OK) {
    return TCL_ERROR;
  }

  if (getGSignalID (interp, type, objv[3], &id) != TCL_OK) {
    cval_free (arg, sig->type, sig->len);
    return TCL_ERROR;
  }

  if (!arg) {
    return TCL_ERROR;
  }

  /* void guint gulong */
  g_signal_remove_emission_hook (id, arg[2].long_val);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::signal connect:
 * g::signal connect
 * g_signal_connect_closure
 * 
 *> Usage
 *   @obj            %GObject*
 *   @signal         string
 *   @script         script
 *   @after          optional %gboolean, defaults to %false
 *   Return value    %gulong connection id
 *
 * Wraps %g_signal_connect_closure. The script is evaluated globally when
 * the signal is triggered. When a signal is triggered, the global variable
 * %_ will be set to a list of the signal arguments.
 **/

static int cmd_g_signal_connect (Tcl_Interp *interp, int objc,
				 Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  int after;

  /* hack to make last argument optional */

  if (objc == 6) {
    /* G_TYPE_OBJECT G_TYPE_STRING G_TYPE_CLOSURE */
    arg = cval_gather (interp, objc, objv, 6, sig, 0x8, "gboolean");
    
    if (Tcl_GetBooleanFromObj (interp, objv[5], &after) != TCL_OK) {
        cval_free (arg, sig->type, sig->len);
	return TCL_ERROR;
    }
  } else {
    /* G_TYPE_OBJECT G_TYPE_STRING G_TYPE_CLOSURE */
    arg = cval_gather (interp, objc, objv, 5, sig, 0);
    after = 0;
  }

  if (!arg) {
    return TCL_ERROR;
  }

  /* gulong GObject* {const gchar*} GClosure* gboolean */
  arg[0].long_val = g_signal_connect_closure (arg[1].ptr_val, arg[2].ptr_val,
					      arg[3].ptr_val, after);

  /* G_TYPE_ULONG */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::type children:
 * g::type children
 * g_type_children
 * 
 *> Usage
 *   @type          %GType
 *   Return value   list of %GType.
 *
 * Wraps %g_type_children.
 **/

static int cmd_g_type_children (Tcl_Interp *interp, int objc,
				Tcl_Obj *CONST objv[],
				const cmd_sig_t *sig)
{
  cval_t *arg;
  Tcl_Obj *obj;
  guint len;
  GType *type;

  /* G_TYPE_UINT */
  arg = cval_gather (interp, objc, objv, 3, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* GType* GType guint* */
   type = g_type_children (arg[1].int_val, &len);

   obj = Tcl_NewListObj (0, 0);

   while (len-- > 0) {
     Tcl_ListObjAppendElement (interp, obj, Tcl_NewIntObj (*type++));
   }

  Tcl_SetObjResult (interp, obj);

  g_free (type);
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gobject.func.g::type interfaces: g::type interfaces
 * g::type interfaces
 * g_type_interfaces
 * 
 *> Usage
 *   @type          %GType
 *   Return value   list of %GType.
 *
 * Wraps %g_type_interfaces.
 *
 **/

static int cmd_g_type_interfaces (Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GType *type;
  guint len;
  Tcl_Obj *obj;

  /* G_TYPE_UINT */
  arg = cval_gather (interp, objc, objv, 3, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* GType* GType guint* */
  type = g_type_interfaces (arg[1].int_val, &len);

  obj = Tcl_NewListObj (0, 0);
  
  while (len-- > 0) {
    Tcl_ListObjAppendElement (interp, obj, Tcl_NewIntObj (*type++));
  }
  
  Tcl_SetObjResult (interp, obj);
  g_free (type);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

