#ifndef TCL_GOBJECT_H
#define TCL_GOBJECT_H

#include <tcl.h>
#include <glib-object.h>

/* Type for strings that have to be freed */

extern GType G_TYPE_STRING_DYNAMIC;

void initHandleType ();

#define HANDLE_PTR(obj) (obj)->internalRep.twoPtrValue.ptr1
#define HANDLE_GTYPE(obj) GPOINTER_TO_UINT((obj)->internalRep.twoPtrValue.ptr2)

/* Return the class of type */
void *type_class (GType type);

/* Commands for operating on handles and cmds in the type system */
int cmd_handle (ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *CONST objv[]);
int cmd_cmd (ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *CONST objv[]);

/* Convesion between Tcl objects and GObject types */

Tcl_Obj *newHandle (void *ptr, GType type);
int getHandle (Tcl_Interp *interp, Tcl_Obj *obj, GType type, void **ptr);

Tcl_Obj *newGEnum (GType type, int val);
int getGEnum (Tcl_Interp *interp, Tcl_Obj *obj, GType type, int *val);

Tcl_Obj *newGFlags (GType type, int val);
int getGFlags (Tcl_Interp *interp, Tcl_Obj *obj, GType type, int *val);

Tcl_Obj *newGValue (const GValue *val);
int getGValue (Tcl_Interp *interp, Tcl_Obj *obj, GValue *val);

Tcl_Obj *newArgv(char **argv);
int getArgv(Tcl_Interp *interp, Tcl_Obj *obj, char ***argv);

GClosure *getGClosure (Tcl_Interp *interp, Tcl_Obj *obj);

/* Set var to be a representation of GError */

int setGErrorVar (Tcl_Interp *interp, Tcl_Obj *var, GError *err);

/* Append strings to the interp result. End with (char *) 0 */

void appendResult (Tcl_Interp *interp, ...);

/* Callback to eval script */

typedef struct cb {
  Tcl_Interp *interp;
  Tcl_Obj *script;
  Tcl_Obj *data;
} cb_t;

cb_t *cb_new (Tcl_Interp *interp, Tcl_Obj *script, Tcl_Obj *data);
void cb_free (cb_t *cb);
Tcl_Obj *cb_exec (cb_t *cb);
void cb_exec_once (cb_t *cb);

/* Values passed to C funcs  */

typedef union _cval {
  char char_val;
  short short_val;
  int int_val;
  long long_val;
  float float_val;
  double double_val;
  void *ptr_val;
} cval_t;

Tcl_Obj *newType (GType type, cval_t *v);
int getType (Tcl_Interp *interp, Tcl_Obj *obj, GType type, cval_t *v);

/* type for GValue must be set */
Tcl_Obj *newGValue (const GValue *val);
int getGValue (Tcl_Interp *interp, Tcl_Obj *obj, GValue *val);

/* For getGValue, val has to have g_value_init done. Don't call
 * g_value_unset on GValue after getGValue, because we do bad hacks.
 * FIXME: investigate this.
 */

typedef void (*generic_func)();
typedef void (*wrap_func) (const generic_func, cval_t *);

/* Signature of a C function.
 * type[0] is the return type.
 * If wrap is NULL, then it is an incomplete sig.
 * Incomplete sigs are used when a command has to be manually implemented.
 * Incomplete sigs have a return type (G_TYPE_INVALID if not understood)
 * and types of some arguments.
*/

typedef struct _sig {
  const wrap_func wrap;    /* Calls a function with right args */
  const GType *type;       /* GTypes of args */
  const guint8 len;        /* Number of args */
} cmd_sig_t;

/* A Tcl command that calls a C function.
 * If sig->wrap is NULL than func is an eval_func because
 * the command had to be manually implemented.
 */

typedef struct _cmd {
  const char *name;         /* Name of the Tcl command */
  const generic_func func;  /* C function called */
  const cmd_sig_t *sig;     /* Sig of the func */
} cmd_t;

typedef int (*eval_func) (Tcl_Interp *interp, int objc,
			  Tcl_Obj *CONST objv[], const cmd_sig_t *sig);

/* Register the cmd with its GType and create a proc for it */

void cmd_register (Tcl_Interp *interp, GType type, const cmd_t *cmd,
		   char *proc);

/* Evaluate a command where clientData is a ptr to cmd_t */

int cmd_eval (ClientData clientData, Tcl_Interp *interp, int objc,
	      Tcl_Obj *CONST objv[]);

/* Evaluate a command where clientData is a ptr to handle Tcl_Obj * by
 * climbing the class heirarchy trying to invoke the command on the handle.
 */

int cmd_obj_eval (ClientData clientData, Tcl_Interp *interp, int objc,
		  Tcl_Obj *CONST objv[]);

/* Return arguments in objv as a cval_t array according to type.
 * n_args is the number of arguments the command takes.
 * Skip indicates where objv has a corresponding type.
 * If bit n (the rightmost bit is bit 0) of skip is 0,
 * objv[n + 2] is converted to the next type. If the bit is 1, objv[n + 2]
 * is skipped. Returns NULL on failure and set an error result in the interp.
 * The cval_t array must be freed with cval_free.
 * For each arg skipped a string must be passed which identifies the
 * argument. This is used on error to set usage info.
 */

cval_t *cval_gather (Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],
		     int n_args, const cmd_sig_t *sig, guint skip, ...);

/* Frees the array returned by cmd_gather. Len and type must be from the
 * sig passed to cval_gather.
 */

void cval_free (cval_t *arg, const GType *type, int len);

#include "wrap.h"

#endif
