# Provide a reasonable default bgerror

proc bgerror {msg} {
    global errorInfo

    puts stderr "background error: $msg"
    puts stderr "stack trace:"
    puts stderr $errorInfo
}

##!
# gobject.util.g::new:
# g::new
#
#> Usage
#   @switches     '-cmd', '-name' @str
#   @typename     name of a %GObject type
#   @arg...       property values
#   Return value  %GObject of type @typename
#
# The '-cmd' switch indicates a new command should be created that
# invokes the new object. The '-name' switch implies '-cmd'.
# If -name is given, the name of the command is str. Otherwise the
# name is that of new object. The command is deleted when the object
# is destroyed. Additionally, if the command was named, a 'handle' option is
# added which returns the wrapped object.
#
# If there is just one @arg, it is taken as a list of property name value
# pairs. Otherwise the list comprising the mulitple @args in the same manner.
# 
#: Examples
#{{
#   set b [g::new GtkButton label test relief half]
#   set b2 [g::new GtkButton [list label test relief half]]
#   g::object set $b label TEST
#   gtk::button set_relief $b2 normal
#
#   set b [g::new -cmd GtkButton label test relief half]
#   $b set use-underline false
#
#   set b [g:::new -name .b GtkButton [list label test relief half]]
#   .b set use-underline false
#   set b [.b handle]
#   g::object set $b use-underline false
#}}
##

proc g::new {args} {
    set cmd 0
    set name ""

    if {[lindex $args 0] eq "-cmd"} {
	set cmd 1
	set args [lrange $args 1 end]
    }

    if {[lindex $args 0] eq "-name"} {
	set cmd 1
	set name [lindex $args 1]
	set args [lrange $args 2 end]
    }

    set n [llength $args]

    if {$n == 0} {
	error "usage: g::new ?-cmd? ?-name str? typename arg..."
    }

    set type [g::type from_name [lindex $args 0]]

    if {$type == 0} {
        error "Unknown type [lindex $args 0]"
    }

    if {$n == 2} {
	set args [lindex $args 1]
    } else {
	set args [lrange $args 1 end]
    }

    set o [g::object new $type $args]

    if {[g::handle is_null $o]} {
        error "failed to create a $type"
    }
    
    if {$cmd} {
	if {$name eq ""} {
	    proc ::$o {cmd args} "g::cmd invoke_object {$o} \$cmd \$args"
	    set name $o
	} else {
	    proc ::$name {cmd args} "if {\$cmd eq \"handle\"} {return $o};\
		    g::cmd invoke_object {$o} \$cmd \$args"
	}

	g::object set_data_full $o _destroy "" [list rename $name ""]
    }

    return $o
}
