#include <tcl-gobject.h>
#include <string.h>

void appendResult (Tcl_Interp *interp, ...)
{
  Tcl_Obj *obj;
  va_list ap;

  va_start (ap, interp);

  obj = Tcl_GetObjResult (interp);
  Tcl_AppendStringsToObjVA (obj, ap);

  va_end(ap);
}

void *type_class (GType type)
{
  void *klass = g_type_class_peek (type);

  if (!klass) {
    klass = g_type_class_ref (type);
  }

  g_assert (klass != 0);

  return klass;
}

int getGEnum (Tcl_Interp *interp, Tcl_Obj *obj, GType type, int *val)
{
  GEnumClass *klass;
  GEnumValue *en;

  klass = type_class (type);

  en = g_enum_get_value_by_nick (klass, Tcl_GetString(obj));

  if (!en) {
    int i;
    
    appendResult (interp, "bad ", g_type_name (type), " \"",
		  Tcl_GetString (obj), "\", must be: ", (char *)0);

    for (i = 0; i < klass->n_values; i++) {
      en = klass->values + i;
      appendResult (interp, en->value_nick,  ", ", (char *)0);
    }

    return TCL_ERROR;
  }

  *val = en->value;
  return TCL_OK;
}

Tcl_Obj *newGEnum (GType type, int val)
{
  GEnumClass *klass;
  GEnumValue *en;

  klass = type_class (type);

  en = g_enum_get_value (klass, val);
  g_assert (en != 0);

  return Tcl_NewStringObj (en->value_nick, -1);
}

/* GFlags string rep is a list of nicks */

int getGFlags (Tcl_Interp *interp, Tcl_Obj * obj, GType type, int *val)
{
  int i, len;
  Tcl_Obj **obj2;
  GFlagsClass *klass;
  GFlagsValue *f;
  
  klass = type_class (type);

  if (Tcl_ListObjGetElements (interp, obj, &len, &obj2) != TCL_OK) {
    return TCL_ERROR;
  }

  *val = 0;

  for (i = 0; i < len; i++) {
    f = g_flags_get_value_by_nick(klass, Tcl_GetString(obj2[i]));
    
    if (!f) {
      appendResult (interp, "bad ", g_type_name(type), ", must be: ", 
		    (char *)0);
     
      for (i = 0; i < klass->n_values; i++) {
	f = klass->values + i;
	appendResult (interp, f->value_nick,  ", ", (char *) 0);
      }
       
      g_type_class_unref (klass);
      return TCL_ERROR;
    }
   
    *val |= f->value;
  }

  return TCL_OK;
}

Tcl_Obj *newGFlags (GType type, int val)
{
  GFlagsClass *klass;
  GFlagsValue *f;
  Tcl_Obj *obj;
  int i;

  obj  = Tcl_NewListObj(0, 0);

  klass = type_class (type);

  for (i = 0; i < klass->n_values; i++) {
    f = klass->values + i;
    
    if (f->value & val) {
      Tcl_ListObjAppendElement(0, obj, Tcl_NewStringObj(f->value_nick, -1));
    }
  }

  return obj;
}

cb_t *cb_new (Tcl_Interp *interp, Tcl_Obj *script, Tcl_Obj *data)
{
  cb_t *cb  = g_new (cb_t, 1);

  cb->interp = interp;
  cb->script = script;
  cb->data = data;

  if (script) {
    Tcl_IncrRefCount (script);
  }

  if (data) {
    Tcl_IncrRefCount (data);
  }
  
  if (cb->interp) {
    Tcl_Preserve (cb->interp);
  }

  return cb;
}

void cb_free (cb_t *cb)
{
  if (cb->script) {
    Tcl_DecrRefCount (cb->script);
  }

  if (cb->data) {
    Tcl_DecrRefCount (cb->data);
  }

  if (cb->interp) {
    Tcl_Release (cb->interp);
  }

  g_free (cb);
}

Tcl_Obj *cb_exec (cb_t *cb)
{
  Tcl_SavedResult state;
  Tcl_Obj *obj;

  Tcl_SaveResult (cb->interp, &state);  

  if (Tcl_EvalObjEx (cb->interp, cb->script, TCL_EVAL_GLOBAL) == TCL_ERROR) {
    Tcl_BackgroundError (cb->interp);
    obj = 0;
  } else {
    obj = Tcl_DuplicateObj (Tcl_GetObjResult (cb->interp));
  }

  Tcl_RestoreResult (cb->interp, &state);

  return obj;
}

void cb_exec_once (cb_t *cb)
{
  cb_exec (cb);
  cb_free (cb);
}

static void cb_free_closure (gpointer data, GClosure *closure)
{
  cb_free (data);
}

static void tcl_closure_marshal (GClosure *closure, GValue *ret_val,
			  guint n_param, const GValue *param,
			  gpointer hint, gpointer data)
{
  int i;
  Tcl_Obj *ret, *arg;
  cb_t *cb = closure->data;

  arg = Tcl_NewListObj (0, 0);

  for (i = 0; i < n_param; i++) {
    Tcl_ListObjAppendElement (cb->interp, arg, newGValue (&param[i]));
  }

  Tcl_SetVar2Ex (cb->interp, "_", 0, arg, TCL_GLOBAL_ONLY);

  ret = cb_exec (cb);

  /* FIXME: some signals with void return values have type set to
   * invalid? */

  if (ret && ret_val && G_VALUE_TYPE (ret_val) != G_TYPE_INVALID) {
    if (getGValue (cb->interp, ret, ret_val) != TCL_OK) {
      Tcl_BackgroundError (cb->interp);
    }
  }
}

int getGValue (Tcl_Interp *interp, Tcl_Obj *obj, GValue *val)
{
  return getType (interp, obj, G_VALUE_TYPE (val), (cval_t *) val->data);
}

Tcl_Obj *newGValue (const GValue *val)
{
  return newType (G_VALUE_TYPE (val), (cval_t *) val->data);
}

GClosure *getGClosure (Tcl_Interp *interp, Tcl_Obj *obj)
{
  GClosure *closure;
  cb_t *cb;
  
  cb = cb_new (interp, obj, 0);
  closure = g_cclosure_new ((GCallback) cb_exec, cb, cb_free_closure);
  g_closure_set_marshal (closure, tcl_closure_marshal);

  return closure;
}

int setGErrorVar (Tcl_Interp *interp, Tcl_Obj *var, GError *err)
{
  Tcl_Obj *obj;

  g_assert (err != 0);

  obj = Tcl_NewListObj (0, 0);

  Tcl_ListObjAppendElement (interp, obj, Tcl_NewIntObj (err->domain));
  Tcl_ListObjAppendElement (interp, obj, Tcl_NewIntObj (err->code));
  Tcl_ListObjAppendElement (interp, obj, Tcl_NewStringObj (err->message, -1));

  obj = Tcl_ObjSetVar2 (interp, var, 0, obj, 0);
  
  if (!obj) {
    return TCL_ERROR;
  }
  
  return TCL_OK;
}

int getArgv(Tcl_Interp *interp, Tcl_Obj *obj, char ***argv)
{
  Tcl_Obj *obj2;
  int i, len;
  char **arg;

  if (Tcl_ListObjLength(interp, obj, &len) != TCL_OK)
    return TCL_ERROR;

  arg = (char **)Tcl_Alloc((len + 1) * sizeof(char *));
 
  for (i = 0; i < len; i++) {
    Tcl_ListObjIndex(0, obj, i, &obj2);

    arg[i] = Tcl_GetString(obj2);
  }
  arg[i] = 0;

  *argv = arg;

  return TCL_OK;
}

Tcl_Obj *newArgv(char **argv)
{
  Tcl_Obj *obj;

  obj = Tcl_NewListObj(0, 0);

  for (; *argv; argv++) {
    Tcl_ListObjAppendElement(0, obj, Tcl_NewStringObj(*argv, -1));
  }

  return obj;
}
