/**!
 * gtk.notes:
 *
 * Almost all the functionality of Gtk is available. Work needs to be done
 * on the new tree and text widgets.
 *
 * New commands were added to create %GtkTextIter and %GtkTreeIter as well
 * as add stock items.
 **/



/**!
 * gtk.func.gtk::text_iter new:
 * gtk::text_iter new
 *
 * Adds the ability to create a %GtkTextIter since it can't be allocated
 * on the heap like in C. Must be freed with %gtk_text_iter_free.
 **/

static GtkTextIter *gtk_text_iter_new ()
{
  GtkTextIter iter;

  return gtk_text_iter_copy (&iter);
}

/**!
 * gtk.func.gtk::tree_iter new:
 * gtk::tree_iter new
 * 
 * Adds the ability to create a %GtkTreeIter since it can't be allocated
 * on the heap like in C. Must be freed with %gtk_tree_iter_free.
 **/

static GtkTreeIter *gtk_tree_iter_new ()
{
  GtkTreeIter iter;

  return gtk_tree_iter_copy (&iter);
}

/**!
 * gtk.func.gtk::toolbar insert_stock:
 * gtk::toolbar insert_stock
 * gtk_toolbar_insert_stock
 *
 *> Usage
 *   @name         %GtkToolbar*
 *   @name         %char*
 *   @name         %char*
 *   @name         %char*
 *   @name         %gint
 *   Return value  %GtkWidget*
 *
 * Wraps %gtk_toolbar_insert_stock. Have to bind own script to 'clicked'.
 **/

static int cmd_gtk_toolbar_insert_stock (Tcl_Interp *interp, int objc,
					 Tcl_Obj *CONST objv[],
					 const cmd_sig_t *sig)
{
  cval_t *arg;

  /* GTK_TYPE_TOOLBAR G_TYPE_STRING G_TYPE_STRING G_TYPE_STRING G_TYPE_INT */
  arg = cval_gather (interp, objc, objv, 7, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* Can't do the callback. No way to destroy it when the widget is
   * destroyed.
   */

  /* GtkWidget* GtkToolbar* {const gchar*} {const char*}
   *     {const char*} GtkSignalFunc gpointer gint
   */

  arg[0].ptr_val = gtk_toolbar_insert_stock (arg[1].ptr_val, arg[2].ptr_val,
					     arg[3].ptr_val, arg[4].ptr_val,
					     0, 0, arg[5].int_val);

  /* GTK_TYPE_WIDGET */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::stock add:
 * gtk::stock add
 * gtk_stock_add_static
 *
 *> Usage
 *   @id       %char*
 *   @label    %char*
 *   @mod      %GdkModifierType
 *   @keyval   %char*
 *
 * Wraps %gtk_stock_add_static. The arguments are the contents of
 * a %GtkStockItem except that @keyval is a string.
 **/

void mod_gtk_stock_add (const char *id, const char *label, GdkModifierType mod,
			const char *keyval)
{
  GtkStockItem item;

  item.stock_id = g_strdup (id);
  item.label = g_strdup (label);
  item.modifier = mod;
  item.keyval = gdk_keyval_from_name (keyval);

  /* FIXME: What is this? */
  item.translation_domain = 0;

  gtk_stock_add (&item, 1);
}

static int getGTypeArray (Tcl_Interp *interp, Tcl_Obj * obj, GType **arr_ptr,
			  int *arr_len)
{
  int i, len;
  Tcl_Obj **obj2;
  GType *arr;

  if (Tcl_ListObjGetElements(interp, obj, &len, &obj2) != TCL_OK) {
    return TCL_ERROR;
  }

  *arr_ptr = arr = g_new (GType, len);
  *arr_len = len;

  for (i = 0; i < len; i++) {
    if (Tcl_GetLongFromObj (interp, obj2[i], arr + i) != TCL_OK) {
      g_free (*arr_ptr);
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}

/**!
 * gtk.func.gtk::tree_store new:
 * gtk::tree_store new
 * gtk_tree_store_newv
 *
 *> Usage
 *   @name          %gint
 *   @name          %GType*
 *   Return value   %GtkTreeStore*
 *
 * Wraps %gtk_tree_store_newv.
 **/

static int cmd_gtk_tree_store_new (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  GtkTreeStore *store;
  GType *type;
  int len;

  if (objc != 3) {
    Tcl_WrongNumArgs (interp, 2, objv, "{type}");
    return TCL_ERROR; 
  }

  if (getGTypeArray (interp, objv[2], &type, &len) != TCL_OK) {
    return TCL_ERROR;
  }

  store = gtk_tree_store_newv (len, type);

  g_free (type);
  Tcl_SetObjResult (interp, newHandle (store, GTK_TYPE_TREE_STORE));
  return TCL_OK;
}

/**!
 * gtk.func.gtk::list_store new:
 * gtk::list_store new
 * gtk_list_store_newv
 *
 *> Usage
 *   @types         list of %GType
 *   Return value   %GtkListStore*
 *
 * Wraps %gtk_list_store_newv.
 **/

static int cmd_gtk_list_store_new (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  GtkListStore *store;
  GType *type;
  int len;

  if (objc != 3) {
    Tcl_WrongNumArgs (interp, 2, objv, "{type}");
    return TCL_ERROR; 
  }

  if (getGTypeArray (interp, objv[2], &type, &len) != TCL_OK) {
    return TCL_ERROR;
  }

  store = gtk_list_store_newv (len, type);

  g_free (type);
  Tcl_SetObjResult (interp, newHandle (store, GTK_TYPE_LIST_STORE));
  return TCL_OK;
}

/**!
 * gtk.func.gtk::message_dialog new:
 * gtk::message_dialog new
 * gtk_message_dialog_new
 *
 *> Usage
 *   @win           %GtkWindow*
 *   @name          %GtkDialogFlags
 *   @name          %GtkMessageType
 *   @name          %GtkButtonsType
 *   @msg           %gchar*
 *   Return value   %GtkWidget*
 *
 * Wraps %gtk_message_dialog_new.
 **/

static int cmd_gtk_message_dialog_new (Tcl_Interp *interp, int objc,
				       Tcl_Obj *CONST objv[],
				       const cmd_sig_t *sig)
{
  cval_t *arg;

  /* GTK_TYPE_WINDOW GTK_TYPE_DIALOG_FLAGS GTK_TYPE_MESSAGE_TYPE GTK_TYPE_BUTTONS_TYPE G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 7, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* GtkWidget* GtkWindow* GtkDialogFlags GtkMessageType GtkButtonsType {const gchar*} ... */
  arg[0].ptr_val = gtk_message_dialog_new (arg[1].ptr_val, arg[2].int_val, arg[3].int_val, arg[4].int_val, "%s", (char *) arg[5].ptr_val);

  /* GTK_TYPE_WIDGET */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::list_store set:
 * gtk::list_store set
 * gtk_list_store_set_value
 *
 *> Usage
 *   @name   %GtkListStore*
 *   @name   %GtkTreeIter*
 *   @name   %gint
 *   @val    any type
 *
 * Wraps %gtk_list_store_set_value.
 **/

static int cmd_gtk_list_store_set (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GValue val = {0};

  /* GTK_TYPE_LIST_STORE GTK_TYPE_TREE_ITER G_TYPE_INT */
  arg = cval_gather (interp, objc, objv, 6, sig, 0X8, "val");

  if (!arg) {
    return TCL_ERROR;
  }

  g_value_init (&val, gtk_tree_model_get_column_type (arg[1].ptr_val,
						     arg[3].int_val));

  if (getGValue (interp, objv[5], &val) != TCL_OK) {
    return TCL_ERROR;
  }

  /* void GtkListStore* GtkTreeIter* gint GValue */
  gtk_list_store_set_value (arg[1].ptr_val, arg[2].ptr_val, arg[3].int_val,
			    &val);

/*   FIXME: g_value_unset (&val); */
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::tree_store set:
 * gtk::tree_store set
 * gtk_tree_store_set_value
 *
 *> Usage
 *   @name   %GtkTreeStore*
 *   @name   %GtkTreeIter*
 *   @name   %gint
 *   @val    any type
 *
 * Wraps %gtk_tree_store_set_value.
 **/


static int cmd_gtk_tree_store_set (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GValue val = {0};

  /* GTK_TYPE_LIST_STORE GTK_TYPE_TREE_ITER G_TYPE_INT */
  arg = cval_gather (interp, objc, objv, 6, sig, 0X8, "val");

  if (!arg) {
    return TCL_ERROR;
  }

  g_value_init (&val, gtk_tree_model_get_column_type (arg[1].ptr_val,
						     arg[3].int_val));

  if (getGValue (interp, objv[5], &val) != TCL_OK) {
    return TCL_ERROR;
  }

  /* void GtkTreeStore* GtkTreeIter* gint GValue* */
  gtk_tree_store_set_value (arg[1].ptr_val, arg[2].ptr_val, arg[3].int_val,
			    &val);

/*    g_value_unset (&val); */
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::tree_model get:
 * gtk::tree_model get
 * gtk_tree_model_get_value
 *
 *> Usage
 *   @name:         %GtkTreeModel*
 *   @name          %GtkTreeIter*
 *   @name          %gint
 *   Return value   val
 *
 * Wraps %gtk_tree_model_get_value.
 **/


static int cmd_gtk_tree_model_get (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GValue val = {0};

  /* GTK_TYPE_TREE_MODEL GTK_TYPE_TREE_ITER G_TYPE_INT */
  arg = cval_gather (interp, objc, objv, 5, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* FIXME
  g_value_init (&val, gtk_tree_model_get_column_type (arg[1].ptr_val,
						      arg[3].int_val));
  */

  /* void GtkTreeModel* GtkTreeIter* gint GValue* */
  gtk_tree_model_get_value (arg[1].ptr_val, arg[2].ptr_val, arg[3].int_val,
			    &val);

  Tcl_SetObjResult (interp, newGValue (&val));
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::tree_selection get_selected:
 * gtk::tree_selection get_selected
 * gtk_tree_selection_get_selected
 *
 *> Usage
 *   @name          %GtkTreeSelection*
 *   @name          %GtkTreeIter*
 *   Return value   %gboolean
 *
 * Wraps %gtk_tree_selection_get_selected.
 **/

static int cmd_gtk_tree_selection_get_selected (Tcl_Interp *interp, int objc,
						Tcl_Obj *CONST objv[],
						const cmd_sig_t *sig)
{
  cval_t *arg;

  /* GTK_TYPE_TREE_SELECTION GTK_TYPE_TREE_ITER */
  arg = cval_gather (interp, objc, objv, 4, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* gboolean GtkTreeSelection* GtkTreeModel** GtkTreeIter* */
  arg[0].int_val = gtk_tree_selection_get_selected (arg[1].ptr_val, 0,
						    arg[2].ptr_val);
  
  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));
  
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::menu popup:
 * gtk::menu popup
 * gtk_menu_popup
 *
 *> Usage
 *   @name   %GtkMenu*
 *   @name   %GtkWidget*
 *   @name   %GtkWidget*
 *   @name   %guint
 *   @name   %guint32
 *
 * Wraps %gtk_menu_popup. No support for menu positioning func.
 **/

static int cmd_gtk_menu_popup (Tcl_Interp *interp, int objc,
			       Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;

  /* GTK_TYPE_MENU GTK_TYPE_WIDGET GTK_TYPE_WIDGET G_TYPE_UINT G_TYPE_UINT */
  arg = cval_gather (interp, objc, objv, 7, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* void GtkMenu* GtkWidget* GtkWidget* GtkMenuPositionFunc gpointer guint guint32 */
  gtk_menu_popup (arg[1].ptr_val, arg[2].ptr_val, arg[3].ptr_val, 0, 0,
		  arg[4].int_val, arg[5].int_val);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gtk.func.gtk::editable insert_text:
 * gtk::editable insert_text
 * gtk_editable_insert_text
 *
 *> Usage
 *   @name          %GtkEditable*
 *   @name          %gchar*
 *   @name          %gint
 *   Return value   %gint new cursor position
 *
 * Wraps %gtk_editable_insert_text.
 **/

static int cmd_gtk_editable_insert_text (Tcl_Interp *interp, int objc,
					 Tcl_Obj *CONST objv[],
					 const cmd_sig_t *sig)
{
  cval_t *arg;

  /* GTK_TYPE_EDITABLE G_TYPE_STRING G_TYPE_INT */
  arg = cval_gather (interp, objc, objv, 5, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* void GtkEditable* {const gchar*} gint gint* */
  gtk_editable_insert_text (arg[1].ptr_val, arg[2].ptr_val,
			    strlen (arg[2].ptr_val), &arg[3].int_val);

  Tcl_SetObjResult (interp, Tcl_NewIntObj (arg[3].int_val));
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}
