# Load pkg.conf in dir

proc pkg_load_conf {dir} {
    global pkg_conf

    array set pkg_conf {Prefix "" PartOf ""}

    set fd [open [file join $dir pkg.conf]] 
    while {[gets $fd line] != -1} {
	if {[string match \#* $line] || [regexp {^\s*$} $line]} {
	    continue
	}

	if {![regexp {^\s*(\w+):\s+(.*?)\s*$} $line x key val]} {
	    error "error parsing \"$line\": must be \"key: val\"."
	}
	
	set pkg_conf($key) $val
    }
    close $fd

    foreach key {Name Id PkgConfigId} {
	if {![info exists pkg_conf($key)]} {
	    error "$key not defined"
 	}
    }

    set pkg_conf(file,map) [file join $dir $pkg_conf(Id).map]
    set pkg_conf(file,def) [file join $dir $pkg_conf(Id).def]
    set pkg_conf(file,pkg) [file join $dir $pkg_conf(Id).pkg]
    set pkg_conf(file,pkg_mod) $pkg_conf(file,pkg).mod
    set pkg_conf(file,stub_c) [file join $dir stub.c]
    set pkg_conf(file,pkg_c) [file join $dir pkg.c]
    set pkg_conf(file,pkg_init) [file join $dir pkg.init]
    set pkg_conf(file,type_def) type.def
    set pkg_conf(file,wrap_def) wrap.def
    set pkg_conf(file,wrap_c) ../src/gobject/wrap.c
    set pkg_conf(file,wrap_h) ../src/gobject/wrap.h

    if {![info exists pkg_conf(IncludeDirs)]} {
	set pkg_conf(IncludeDirs) [pkg_include_dirs]
    }
}

# Figure out where the include files are using pkg-config

proc has_headers {dir} {
    return [llength [glob -nocomplain -directory $dir *.h]]
}

proc pkg_header_path {header}  {
    global pkg_conf

    foreach dir $pkg_conf(IncludeDirs) {
	set test [file join $dir $header]
	if {[file exists $test]} {
	    return $test
	}
    }

    error "could not find $header for package $pkg_confName) in\
          $pkg_conf(IncludeDirs)"
}

# FIXME: problems with gnome having matching two many include dirs

proc pkg_include_dirs {} {
    global pkg_conf

    set cflags [exec pkg-config --cflags $pkg_conf(PkgConfigId)]
    set dirs [list]

    set maybe_dirs [regexp -all -inline -- {-I\s*(\S+)} $cflags]
    lappend maybe_dirs /usr/include

    foreach dir $maybe_dirs {
	if {![file isdir $dir]} {
	    continue
	}

	if {[regexp $pkg_conf(Id)|$pkg_conf(PkgConfigId) $dir]} {
	    if {[has_headers $dir]} {
		lappend dirs $dir
	    }
	}

	foreach id [list $pkg_conf(Id) $pkg_conf(PkgConfigId)] {
	    foreach test [glob -nocomplain -types d -directory $dir *$id*] {
		if {[has_headers $test]} {
		    lappend dirs $test
		}
	    }
	}
    }

    if {[llength $dirs] == 0} {
	error "could not find include directory for $pkg_conf(Name)"
    }

    return [lsort -unique $dirs]
}

# Return all function defined in a header file as a list
# of the form {name return_type arg_type...}

proc cfuncs {header} {
    set fd [open $header r]
    set buf [read $fd [file size $header]]
    close $fd

    set funcs [list]

    # make line continuations one line
    regsub -all {\\\n} $buf "" buf

    # remove comments
    regsub -all {/\*.*?\*/} $buf "" buf

    # remove deprecated funcs
    regsub -all {#ifndef[^\n]*?DISABLE_DEPRECATED.*?#endif} $buf "" buf

    # remove preprocessor commands
    regsub -all {\#[^\n]*\n} $buf "" buf

    # remove all typdefs
    regsub -all {typedef.*?;} $buf "" buf

    # remove extern "C" declaration
    regsub -all {extern\s*?"C"\s*?\{} $buf "" buf

    # remove other type mods
    regsub -all {G_GNUC_CONST|unsigned} $buf "" buf

    # set this to const, is that right?
    regsub -all G_CONST_RETURN $buf const buf

    # delete empty lines
    regsub -all {\n\s*?\n} $buf "" buf

    # misc crap
    regsub -all {\w+_DECLS} $buf "" buf

    # remove structures
    regsub -all {struct.*?\{.*?\};} $buf "" buf

    # Add space after * connected to a func name
    regsub -all {\*([a-z])} $buf {* \1} buf

    #puts stderr [string toupper $header]
    #puts stderr $buf
    #puts stderr ENDENDENDENDEND

    foreach def [split $buf \;] {
	regsub -all {\s+\*} $def "* " def
	if {[regexp {\s*([^\n]+?)\s+(\S+)\s*\(([^\)]*)\)\s*} $def x \
		ret name args]} {
	    regsub -all {\s+\*} $args "* " args
	    set func [list]
	    lappend func $name $ret

	    set arglist [list]
	    foreach el [split $args ,] {
		if {[string match {*\.\.\.*} $el]} {
		    set type ...
		} else {
		    regexp {^(.*?)\S*$} $el x type
		    set type [string trim $type]
		    if {$type eq ""} {
			set type void
		    }
		}

		lappend func $type
	    }
	    
	    lappend funcs $func
	}
    }

    return $funcs
}

# Parse a package into the global var pkg. If pkg already exists
# merge into pkg.
#
# Elements of the pkg array:
#   classes - list of class names
#   class,$class,method - list of form name cfunc ctype gtype...

proc pkg_parse {file} {
    global pkg

    set in [open $file r]

    set pkg_def [read $in]

    regsub -all -lineanchor {^\s*\#[^\n]*} $pkg_def "" pkg_def

    if {![info exists pkg(classes)]} {
	set pkg(classes) [list]
    }

    foreach {statement class method_defs} $pkg_def {
	if {$statement ne "class"} {
	    error "expected class def, not $statement"
	}
	
	if {[lsearch $pkg(classes) $class] == -1} {
	    lappend pkg(classes) $class
	}

	if {![info exists pkg(class,$class,method)]} {
	    set pkg(class,$class,method) [list]
	}

	foreach {statement method method_def} $method_defs {
	    if {$statement ne "method"} {
		error "expected method def, not $statement"
	    }
	    
	    set cfunc ""
	    set ctype ""
	    set gtype ""

	    foreach  {attrib value} $method_def {
		set $attrib $value

		if {![regexp {^(cfunc|ctype|gtype)$} $attrib]} {
		    error "bad method attribute $attrib"
		}
	    }

	    if {$cfunc eq "" || $ctype eq "" || $gtype eq ""} {
		error "ctype, cfunc, and gtype must be given"
	    }
 
	    # Remove G_TYPE_NONE indicating no args

	    if {[lindex $gtype 1] eq "G_TYPE_NONE"} {
		set gtype [lindex $gtype 0]
	    }

	    set x class,$class,method

	    if {![info exists pkg($x)]} {
		set i -1
	    } else {
		# if we're renaming a function remove any existing method
		# using that function

		set i [lsearch $pkg($x) $cfunc]

		if {$i != -1} {
		    set pkg($x) [lreplace $pkg($x) [incr i -1] [incr i 3]]
		}


		set i [lsearch $pkg($x) $method]
	    }

	    if {$i == -1} {
		lappend pkg($x) $method $cfunc $ctype $gtype
	    } else {
		set pkg($x) [lreplace $pkg($x) $i [incr i 3] \
				 $method $cfunc $ctype $gtype]
	    }
	}
    }
}

proc pkg_load {} {
    global pkg_conf pkg

    pkg_parse $pkg_conf(file,pkg)

    if {[file exists $pkg_conf(file,pkg_mod)]} {
	pkg_parse $pkg_conf(file,pkg_mod)
    }
}

# Turn args like GtkButton to gtk_button

proc pkg_id {args} {
    regsub -all {([^_])([A-Z])} [join $args _] {\1_\2} res
    return [string tolower $res]
}

# Return Tcl command name for a class in a package.

proc pkg_cmd {pkg c} {
    return [pkg_id $pkg]::[pkg_id $c]
}

# Load in typemap and create arrays that map from the C type to a
# GType and base type

proc pkg_load_typemap {} {
    global pkg_gtype pkg_btype

    set fd [open typemap r]

    while {[gets $fd line] != -1} {
	set line [string trim $line]

	if {$line eq "" || [string match #* $line]} {
	    continue
	}

	if {[catch {llength $line} n]} {
	    error "badly formed line $line in typemap"
	} 

	if {$n != 3} {
	    error "$line, must be \"ctype gtype base_type\""
	}

	foreach {ct gt bt} $line break

	set pkg_gtype($ct) $gt
	set pkg_btype($ct) $bt
    }

    close $fd
}


# Return code for accessing an array arg of cval_t at index i.

proc pkg_cval {i bt} {
    if {$bt eq "void"} {
	return ""
    } else {
	if {[string match int* $bt]} {
	    set bt int
	}
	
	set field ${bt}_val
    }
    
    return  arg\[${i}\].$field
}


# Return C code to call the func with right arguments and handling return

proc pkg_cval_func {func bt} {
    if {[lindex $bt 0] eq "void"} {
	append code "$func ("
    } elseif {[lindex $bt 0] eq "???"} {
	append code "/* [pkg_cval 0 [lindex $bt 0]] = */ $func ("
    } else {
	append code "[pkg_cval 0 [lindex $bt 0]] = $func ("
    }

    set bt [lrange $bt 1 end]
    set i 0
    set len [llength $bt]

    foreach b $bt {
	if {$b ne "???"} {
	    append code [pkg_cval [incr i] $b]
	} else {
	    append code 0
	}

	if {[incr len -1] > 0} {
	    append code ", "
	}
    }
    
    return [append code ");"]
}
